"""Environment variables used by weave."""

# TODO: we should put all other env vars here to keep them organized.

import configparser
import enum
import os
import json
import typing
from . import util
from . import errors
from urllib.parse import urlparse
import netrc
from distutils.util import strtobool

if typing.TYPE_CHECKING:
    from . import logs

WANDB_ERROR_REPORTING = "WANDB_ERROR_REPORTING"
WEAVE_USAGE_ANALYTICS = "WEAVE_USAGE_ANALYTICS"
WEAVE_GQL_SCHEMA_PATH = "WEAVE_GQL_SCHEMA_PATH"


def _env_as_bool(var: str, default: typing.Optional[str] = None) -> bool:
    env = os.environ
    val = env.get(var, default)
    try:
        val = bool(strtobool(val))  # type: ignore
    except (AttributeError, ValueError):
        pass
    return val if isinstance(val, bool) else False


class Settings:
    """A minimal readonly implementation of wandb/old/settings.py for reading settings"""

    DEFAULT_SECTION = "default"
    DEFAULT_BASE_URL = "https://api.wandb.ai"

    def __init__(self) -> None:
        self._settings = configparser.ConfigParser()
        if not self._settings.has_section(self.DEFAULT_SECTION):
            self._settings.add_section(self.DEFAULT_SECTION)
        self._settings.read([Settings._global_path(), Settings._local_path()])

    @property
    def base_url(self) -> str:
        try:
            return self._settings.get(self.DEFAULT_SECTION, "base_url")
        except configparser.NoOptionError:
            return self.DEFAULT_BASE_URL

    @staticmethod
    def _local_path() -> str:
        return os.path.join(os.getcwd(), "wandb", "settings")

    @staticmethod
    def _global_path() -> str:
        default_config_dir = os.path.join(os.path.expanduser("~"), ".config", "wandb")
        config_dir = os.environ.get("WANDB_CONFIG_DIR", default_config_dir)
        return os.path.join(config_dir, "settings")


# There are currently two cache modes:
# - full: cache all cacheable intermediate results
# - minimal: cache only what we're sure we need to cache for performance
# WEAVE_NO_CACHE indicates that we're using the minimal cache mode. Otherwise
# we'll use full.
class CacheMode(enum.Enum):
    FULL = "full"
    MINIMAL = "minimal"


def cache_mode() -> CacheMode:
    if util.parse_boolean_env_var("WEAVE_NO_CACHE"):
        return CacheMode.MINIMAL
    env_mode = os.getenv("WEAVE_CACHE_MODE", CacheMode.MINIMAL.value)
    for mode in CacheMode:
        if mode.value == env_mode:
            return mode
    raise errors.WeaveConfigurationError(
        f"WEAVE_CACHE_MODE must be one of {list(CacheMode)}"
    )


# represents the number of days to re-use the weave cache before switching to a new directory
# a low number will cause frequent cache rotations and may lead to performance degradation due to
# cache misses. A high number will result in a large cache size which may cause infrastructure issues.
# if unset, the cache will write to a path that will not be cleaned up by the default cleanup task
def cache_duration_days() -> int:
    return int(os.getenv("WEAVE_CACHE_DURATION_DAYS", 0))


# represents the number of days after a cache has expired to leave it until deletion
def cache_deletion_buffer_days() -> int:
    return int(os.getenv("WEAVE_CACHE_DELETION_BUFFER_DAYS", 1))


def wandb_production() -> bool:
    return os.getenv("WEAVE_ENV") == "wandb_production"


def is_public() -> bool:
    return wandb_production()


def weave_log_format(default: "logs.LogFormat") -> "logs.LogFormat":
    from .logs import LogFormat

    return LogFormat(os.getenv("WEAVE_LOG_FORMAT", default))


def weave_link_prefix() -> str:
    """When running in server we mount index under /weave"""
    if os.getenv("GORILLA_ONPREM") == "true":
        return "/weave"
    return ""


def weave_onprem() -> bool:
    return os.getenv("GORILLA_ONPREM") == "true"


def weave_backend_host() -> str:
    return os.getenv("WEAVE_BACKEND_HOST", "/__weave")


def analytics_disabled() -> bool:
    if os.getenv("WEAVE_DISABLE_ANALYTICS") == "true":
        return True
    return False


def weave_server_url() -> str:
    base_url = wandb_base_url()
    default = "https://weave.wandb.ai"
    if base_url != "https://api.wandb.ai":
        default = base_url + "/weave"
    return os.getenv("WEAVE_SERVER_URL", default)


def wandb_base_url() -> str:
    settings = Settings()
    return os.environ.get("WANDB_BASE_URL", settings.base_url).rstrip("/")


def weave_filesystem_dir() -> str:
    # WEAVE_LOCAL_ARTIFACT_DIR should be renamed to WEAVE_FILESYSTEM_DIR
    # TODO
    return os.environ.get("WEAVE_LOCAL_ARTIFACT_DIR") or os.path.join(
        "/tmp", "weave", "fs"
    )


def enable_touch_on_read() -> bool:
    return util.parse_boolean_env_var("WEAVE_ENABLE_TOUCH_ON_READ")


def memdump_sighandler_enabled() -> bool:
    return util.parse_boolean_env_var("WEAVE_ENABLE_MEMDUMP_SIGHANDLER")


def sigterm_sighandler_enabled() -> bool:
    return util.parse_boolean_env_var("WEAVE_ENABLE_SIGTERM_SIGHANDLER")


def weave_wandb_gql_headers() -> typing.Dict[str, str]:
    # expects a json string
    raw = os.environ.get("WEAVE_WANDB_GQL_HEADERS")
    if raw:
        try:
            return json.loads(raw)
        except json.JSONDecodeError:
            raise errors.WeaveConfigurationError(
                "WEAVE_WANDB_GQL_HEADERS should be a json string"
            )
    return {}


def weave_wandb_cookie() -> typing.Optional[str]:
    cookie = os.environ.get("WEAVE_WANDB_COOKIE")
    if cookie:
        if is_public():
            raise errors.WeaveConfigurationError(
                "WEAVE_WANDB_COOKIE should not be set in public mode."
            )
        for netrc_file in ("~/.netrc", "~/_netrc"):
            if os.path.exists(os.path.expanduser(netrc_file)):
                raise errors.WeaveConfigurationError(
                    f"Please delete {netrc_file} while using WEAVE_WANDB_COOKIE to avoid using your credentials"
                )
    return cookie


def stack_dump_sighandler_enabled() -> bool:
    return util.parse_boolean_env_var("WEAVE_ENABLE_STACK_DUMP_SIGHANDLER")


def _wandb_api_key_via_env() -> typing.Optional[str]:
    api_key = os.environ.get("WANDB_API_KEY")
    if api_key and is_public():
        raise errors.WeaveConfigurationError(
            "WANDB_API_KEY should not be set in public mode."
        )
    return api_key


def _wandb_api_key_via_netrc() -> typing.Optional[str]:
    netrc_path = os.path.expanduser("~/.netrc")
    if not os.path.exists(netrc_path):
        return None
    nrc = netrc.netrc(netrc_path)
    res = nrc.authenticators(urlparse(wandb_base_url()).netloc)
    api_key = None
    if res:
        user, account, api_key = res
    if api_key and is_public():
        raise errors.WeaveConfigurationError(
            "~/.netrc should not be set in public mode."
        )
    return api_key


def weave_wandb_api_key() -> typing.Optional[str]:
    env_api_key = _wandb_api_key_via_env()
    netrc_api_key = _wandb_api_key_via_netrc()
    if env_api_key and netrc_api_key:
        raise errors.WeaveConfigurationError(
            "WANDB_API_KEY should not be set in both ~/.netrc and the environment."
        )
    return env_api_key or netrc_api_key


def projection_timeout_sec() -> typing.Optional[typing.Union[int, float]]:
    return util.parse_number_env_var("WEAVE_PROJECTION_TIMEOUT_SEC")


def num_gql_timeout_retries() -> int:
    raw = util.parse_number_env_var("WEAVE_WANDB_GQL_NUM_TIMEOUT_RETRIES")
    if raw is None:
        return 0
    return int(raw)


def usage_analytics_enabled() -> bool:
    return _env_as_bool(WANDB_ERROR_REPORTING, default="True") and _env_as_bool(
        WEAVE_USAGE_ANALYTICS, default="True"
    )


def gql_schema_path() -> typing.Optional[str]:
    return os.environ.get(WEAVE_GQL_SCHEMA_PATH) or None


def dd_env() -> str:
    return os.getenv("DD_ENV", "")


def env_is_ci() -> bool:
    return util.parse_boolean_env_var("WEAVE_CI")



from .. import context_state as _context_state

from .. import logs

# If there is no WEAVE_LOG_LEVEL, this will set to ERROR
# from python's default which is WARNING
logs.configure_logger()

from .dataset import *
from .model import *
from .chat_model import *
from .structured_output import *

from .evaluate import *

from .openai import *
from .anyscale import *
from .huggingface import *

from .faiss import *



import dataclasses
import inspect
import typing

import weave

from ... import weave_internal

from . import weave_plotly


# This is the panel's config (the state that is stored in the panel and configurable
# in the UI by clicking on the gear icon.)
@weave.type()
class GeoConfig:
    x_fn: weave.Node[float] = dataclasses.field(
        default_factory=lambda: weave.graph.VoidNode()
    )
    y_fn: weave.Node[float] = dataclasses.field(
        default_factory=lambda: weave.graph.VoidNode()
    )
    color_fn: weave.Node[float] = dataclasses.field(
        default_factory=lambda: weave.graph.VoidNode()
    )


# This is boilerplate that I'd like to get rid of.
def _geo_set_default_config(config, input_node, new_config):
    return new_config


def _geo_default_config_output_type(input_type):
    # We have to do shenaningans to pass the input type through
    # because it might be a subtype since the input type has unions
    # TOOD: Fix
    config_type = input_type["config"]
    if config_type == None:
        return GeoConfig.WeaveType()
    if isinstance(config_type, weave.types.Const):
        return config_type.val_type
    return config_type


# TODO: really annoying that I need the setter here.
@weave.op(
    setter=_geo_set_default_config,
    output_type=_geo_default_config_output_type,
)
def geo_default_config(
    config: typing.Optional[GeoConfig],
    unnested_node: list[typing.Any],
):
    input_type_item_type = weave.type_of(unnested_node).object_type  # type: ignore
    if config == None:
        return GeoConfig(
            x_fn=weave_internal.define_fn(
                {"item": input_type_item_type}, lambda item: item
            ),
            y_fn=weave_internal.define_fn(
                {"item": input_type_item_type}, lambda item: item
            ),
            color_fn=weave_internal.define_fn(
                {"item": input_type_item_type}, lambda item: item
            ),
        )
    return config


# The render op. This renders the panel.
@weave.op(name="Geo")
def geo(
    input_node: weave.Node[list[typing.Any]], config: GeoConfig
) -> weave_plotly.PanelPlotly:
    unnested = weave.ops.unnest(input_node)
    config = geo_default_config(config, unnested)
    plot_data = unnested.map(
        lambda item: weave.ops.dict_(
            long=config.x_fn(item),  # type: ignore
            lat=config.y_fn(item),  # type: ignore
            color=config.color_fn(item),  # type: ignore
        )
    )
    fig = weave_plotly.plotly_geo(plot_data)
    return weave_plotly.PanelPlotly(fig)


# The config render op. This renders the config editor.
@weave.op(name="Geo_config")
def geo_config(
    input_node: weave.Node[list[typing.Any]], config: GeoConfig
) -> weave.panels.Group:
    unnested = weave.ops.unnest(input_node)
    config = geo_default_config(config, unnested)
    return weave.panels.Group(
        items={
            "x_fn": weave.panels.LabeledItem(
                label="x",
                item=weave.panels.FunctionEditor(
                    config=weave.panels.FunctionEditorConfig(config.x_fn)
                ),
            ),
            "y_fn": weave.panels.LabeledItem(
                label="y",
                item=weave.panels.FunctionEditor(
                    config=weave.panels.FunctionEditorConfig(config.y_fn)
                ),
            ),
            "color_fn": weave.panels.LabeledItem(
                label="color",
                item=weave.panels.FunctionEditor(
                    config=weave.panels.FunctionEditorConfig(config.color_fn)
                ),
            ),
        }
    )


# The interface for constructing this Panel from Python
@weave.type()
class Geo(weave.Panel):
    id = "Geo"
    config: typing.Optional[GeoConfig] = None
    _renderAsPanel: typing.Optional[weave_plotly.PanelPlotly] = dataclasses.field(
        default_factory=lambda: None
    )

    def __init__(
        self, input_node, vars=None, config=None, _renderAsPanel=None, **options
    ):
        super().__init__(input_node=input_node, vars=vars)
        self._renderAsPanel = _renderAsPanel
        if self._renderAsPanel is None:
            self._renderAsPanel = weave_plotly.PanelPlotly()

        self.config = config
        if self.config is None:
            self.config = GeoConfig()

            unnested = weave.ops.unnest(self.input_node)
            if "x_fn" in options:
                sig = inspect.signature(options["x_fn"])
                param_name = list(sig.parameters.values())[0].name
                self.config.x_fn = weave_internal.define_fn(
                    {param_name: unnested.type.object_type}, options["x_fn"]
                )
            else:
                self.config.x_fn = weave_internal.define_fn(
                    {"item": unnested.type.object_type}, lambda item: item
                )
            if "y_fn" in options:
                sig = inspect.signature(options["y_fn"])
                param_name = list(sig.parameters.values())[0].name
                self.config.y_fn = weave_internal.define_fn(
                    {param_name: unnested.type.object_type}, options["y_fn"]
                )
            else:
                self.config.y_fn = weave_internal.define_fn(
                    {"item": unnested.type.object_type}, lambda item: item
                )
            if "color_fn" in options:
                sig = inspect.signature(options["color_fn"])
                param_name = list(sig.parameters.values())[0].name
                self.config.color_fn = weave_internal.define_fn(
                    {param_name: unnested.type.object_type}, options["color_fn"]
                )
            else:
                self.config.color_fn = weave_internal.define_fn(
                    {"item": unnested.type.object_type}, lambda item: item
                )

    # This function currently requires a paired output_type implementation in WeaveJS!
    # TODO: Fix
    @weave.op(output_type=lambda input_type: input_type["self"].input_node.output_type)
    def selected(self):
        # TODO: This function is not right! We need to do a range selection in polar space!
        unnested = weave.ops.unnest(self.input_node)
        config = geo_default_config(self.config, unnested)
        filtered = unnested.filter(
            lambda item: weave.ops.Boolean.bool_and(
                weave.ops.Boolean.bool_and(
                    weave.ops.Boolean.bool_and(
                        config.x_fn(item)
                        > weave.ops.TypedDict.pick(
                            self._renderAsPanel.config.selected, "xMin"
                        ),
                        config.x_fn(item)
                        < weave.ops.TypedDict.pick(
                            self._renderAsPanel.config.selected, "xMax"
                        ),
                    ),
                    config.y_fn(item)
                    > weave.ops.TypedDict.pick(
                        self._renderAsPanel.config.selected, "yMin"
                    ),
                ),
                config.y_fn(item)
                < weave.ops.TypedDict.pick(self._renderAsPanel.config.selected, "yMax"),
            )
        )
        return weave_internal.use(filtered)



import typing
import dataclasses
import transformers

import weave

from . import hfmodel


# We have to forward-declare the Weave types to avoid circular reference
# issues that weave.type() can't resolve yet.


class HFModelTextClassificationType(hfmodel.HFModelType):
    pass


class FullTextClassificationPipelineOutputType(hfmodel.FullPipelineOutputType):
    def property_types(self):
        return {
            "_model": HFModelTextClassificationType(),
            "model_input": weave.types.String(),
            "model_output": weave.types.List(
                weave.types.TypedDict(
                    {
                        "score": weave.types.Float(),
                        "label": weave.types.String(),
                    }
                )
            ),
        }


class ClassificationResultType(weave.types.ObjectType):
    def property_types(self):
        return {
            "_model_name": weave.types.String(),
            "model_input": weave.types.String(),
            "_score": weave.types.Float(),
            "_label": weave.types.String(),
        }


@weave.weave_class(weave_type=ClassificationResultType)
@dataclasses.dataclass
class ClassificationResult:
    _model_name: str
    model_input: str
    _score: float
    _label: str

    @weave.op()
    def model_name(self) -> str:
        return self._model_name

    @weave.op()
    def score(self) -> float:
        return self._score

    @weave.op()
    def label(self) -> str:
        return self._label


ClassificationResultType.instance_classes = ClassificationResult


@weave.type()
class ClassificationResultPanel(weave.Panel):
    id = "ClassificationResultPanel"
    input_node: weave.Node[list[ClassificationResult]]

    @weave.op()
    def render(self) -> weave.panels.Table:
        from .huggingface_models import huggingface

        return weave.panels.Table(
            self.input_node,
            columns=[
                lambda result_row: weave.panels.WeaveLink(
                    result_row.model_name(), to=lambda input: huggingface().model(input)  # type: ignore
                ),
                lambda result_row: result_row.model_input,
                lambda result_row: result_row.score(),
                lambda result_row: result_row.label(),
            ],
        )


class TextClassificationPipelineOutput(typing.TypedDict):
    label: str
    score: float


@weave.weave_class(weave_type=FullTextClassificationPipelineOutputType)
@dataclasses.dataclass
class FullTextClassificationPipelineOutput(hfmodel.FullPipelineOutput):
    _model: "HFModelTextClassification"
    model_input: str
    model_output: list[TextClassificationPipelineOutput]

    @weave.op()
    def model_name(self) -> str:
        return weave.use(self._model.id())


FullTextClassificationPipelineOutputType.instance_classes = (
    FullTextClassificationPipelineOutput
)


@weave.type()
class FullTextClassificationResultPanel(weave.Panel):
    id = "FullTextClassificationResultPanel"
    input_node: weave.Node[FullTextClassificationPipelineOutput]

    @weave.op()
    def render(self) -> weave.panels.Group:
        output = typing.cast(FullTextClassificationPipelineOutput, self.input_node)
        return weave.panels.Group(
            preferHorizontal=True,
            items={
                "input": weave.panels.LabeledItem(
                    label="input", item=output.model_input
                ),
                "output": weave.panels.LabeledItem(
                    label="output",
                    item=weave.panels.Plot(
                        input_node=typing.cast(weave.Node, output.model_output),
                        x=lambda class_score: class_score["score"],
                        y=lambda class_score: class_score["label"],
                    ),
                ),
            },
        )


@weave.weave_class(weave_type=HFModelTextClassificationType)
@dataclasses.dataclass
class HFModelTextClassification(hfmodel.HFModel):
    @weave.op()
    def pipeline(
        self,
    ) -> transformers.pipelines.Pipeline:
        return transformers.pipeline(
            self._pipeline_tag,
            model=self._id,
            return_all_scores=True,
        )

    @weave.op()
    def call(self, input: str) -> FullTextClassificationPipelineOutput:
        output = weave.use(self.pipeline())(input)
        return FullTextClassificationPipelineOutput(self, input, output)

    def _call_list(
        self, input: typing.List[str]
    ) -> typing.List[FullTextClassificationPipelineOutput]:
        output = list(map(weave.use(self.pipeline()), input))
        return [
            FullTextClassificationPipelineOutput(self, i, o)
            for (i, o) in zip(input, output)
        ]

    @weave.op()
    def call_list(self, input: list[str]) -> list[FullTextClassificationPipelineOutput]:
        return self._call_list(input)


@weave.op()
def apply_models(
    models: list[HFModelTextClassification], inputs: list[str]
) -> list[ClassificationResult]:
    results = [model._call_list(inputs) for model in models]
    retval: list[ClassificationResult] = []
    for i, result in enumerate(results):
        for item in result:
            for output in item.model_output:
                retval.append(
                    ClassificationResult(
                        _model_name=models[i]._id,
                        model_input=item.model_input,
                        _score=output["score"],
                        _label=output["label"],
                    )
                )
    return retval


HFModelTextClassificationType.instance_classes = HFModelTextClassification



from concurrent.futures import ThreadPoolExecutor
import contextlib
import contextvars
from typing import Optional, Callable, TypeVar, Iterator, Generator

from . import context
from . import cache
from . import context_state
from . import graph_client_context
from . import run_context
from . import execute
from . import forward_graph
from . import memo
from . import wandb_api

# Must be power of 2
MAX_PARALLELISM = 16
assert MAX_PARALLELISM & (MAX_PARALLELISM - 1) == 0

_parallel_budget_ctx: contextvars.ContextVar[Optional[int]] = contextvars.ContextVar(
    "_parallel_budget_ctx", default=MAX_PARALLELISM
)


def get_parallel_budget() -> int:
    budget = _parallel_budget_ctx.get()
    if budget is None:
        return MAX_PARALLELISM
    return budget


@contextlib.contextmanager
def parallel_budget_ctx(budget: Optional[int]) -> Generator[None, None, None]:
    token = _parallel_budget_ctx.set(budget)
    try:
        yield
    finally:
        _parallel_budget_ctx.reset(token)


ItemType = TypeVar("ItemType")
ResultType = TypeVar("ResultType")


def do_in_parallel(
    do_one: Callable[[ItemType], ResultType], items: list[ItemType]
) -> Iterator[ResultType]:
    parallel_budget = get_parallel_budget()

    if parallel_budget <= 1:
        return map(do_one, items)

    # Contexts aren't automatically propagated to threads, so we have to do so manually for every context
    memo_ctx = memo._memo_storage.get()
    remaining_budget_per_thread = get_remaining_budget_per_thread(len(items))
    wandb_api_ctx = wandb_api.get_wandb_api_context()
    result_store = forward_graph.get_node_result_store()
    top_level_stats = execute.get_top_level_stats()
    eager_mode = context_state.eager_mode()
    graph_client = graph_client_context.get_graph_client()
    run_stack = run_context.get_run_stack()
    cache_prefix = cache.get_cache_prefix_context()

    def do_one_with_memo_and_parallel_budget(x: ItemType) -> ResultType:
        memo_token = memo._memo_storage.set(memo_ctx)
        thread_result_store = None
        thread_top_level_stats = None
        try:
            with parallel_budget_ctx(remaining_budget_per_thread):
                with graph_client_context.set_graph_client(graph_client):
                    with run_context.set_run_stack(run_stack):
                        with context_state.set_eager_mode(eager_mode):
                            with wandb_api.wandb_api_context(wandb_api_ctx):
                                with cache.time_interval_cache_prefix(cache_prefix):
                                    with context.execution_client():
                                        with forward_graph.node_result_store(
                                            result_store
                                        ) as thread_result_store:
                                            with execute.top_level_stats() as thread_top_level_stats:
                                                return do_one(x)
        finally:
            memo._memo_storage.reset(memo_token)
            if thread_result_store is not None:
                result_store.merge(thread_result_store)
            if top_level_stats is not None and thread_top_level_stats is not None:
                top_level_stats.merge(thread_top_level_stats)

    return ThreadPoolExecutor(max_workers=parallel_budget).map(
        do_one_with_memo_and_parallel_budget, items
    )


def get_remaining_budget_per_thread(item_count: int) -> int:
    parallel_budget = get_parallel_budget()
    if item_count <= 0:
        return parallel_budget
    return max(parallel_budget // item_count, 1)



import numpy as np
import matplotlib
import matplotlib.pyplot as plt
import io
from PIL import Image
import weave

matplotlib.use("agg")


@weave.op(render_info={"type": "function"})
def histogram(
    val_series: list[list[float]],
    bin_size: float = 1.0,
    chart_title: str = "Composite Histogram",
    series_names: list[str] = [],
) -> Image.Image:
    """Generate a composite histogram from one or more series of float values provided, using the same bins.
    Return a static image of the result"""
    # TODO: make an even better default x-axis?
    # TODO: check that this will work for e.g. runs.history.accuracy
    #       "displaying the first 10 runs"  in weave.js, vals needs to be at least 2 and no more than 10
    # wouldn't be TYPE ASSIGNABLE
    # TODO: render as a model card/pretty panel...?
    # compute range for values

    fig, ax = plt.subplots(figsize=(15, 5))

    min_val = np.min([np.min(val_list) for val_list in val_series])
    max_val = np.max([np.max(val_list) for val_list in val_series])

    # generate bins across the full value range, where each bin has size bin_size
    bins = np.arange(min_val - bin_size, max_val + bin_size, step=bin_size)
    num_bins = len(bins)

    kwargs = {"edgecolor": "black"}

    num_series = len(val_series)
    # if no names are provided for the series, name each according to its index in the list
    if not series_names:
        series_names = [str(i) for i in range(num_series)]

    # improve color settings for the base cases of 1 or 2 series
    # TODO: I tried manual tuning for 3 series, r/c/b and r/b/m are both not great without hover interaction
    default_colors = {1: ["b"], 2: ["r", "b"]}
    # distribute series evenly across the hue spectrum
    color_map = plt.get_cmap("hsv")
    scale_color = lambda val_id: val_id / len(val_series)

    for series_id, vals in enumerate(val_series):
        curr_color = (
            default_colors[num_series][series_id]
            if num_series < 3
            else color_map(scale_color(series_id))
        )
        ax.hist(
            vals,
            bins,
            alpha=0.5,
            label=series_names[series_id],
            color=curr_color,
            **kwargs
        )

    ax.set_xticks(np.arange(min_val - bin_size, max_val + bin_size, bin_size))
    ax.set_xlim([min_val - bin_size, max_val + bin_size])
    # TODO: may want to improve y-axis as well?
    ax.legend()

    ax.set_xlabel("Values Across Series")
    ax.set_ylabel("Value Count")
    ax.set_title(chart_title)

    img_buf = io.BytesIO()
    fig.savefig(img_buf, format="png")
    im = Image.open(img_buf)
    plt.close(fig)
    return im
